'' Program to determine the best one- or two-tier decomposition of a 16-bit constant (fractional) multiply operation.
'' 
''Copyright (C) 2006 Philip C. Pilgrim
'' 
''This program is free software; you can redistribute it and/or modify
''it under the terms of the GNU General Public License as published by
''the Free Software Foundation; either version 2 of the License, or
''(at your option) any later version.
''
''This program is distributed in the hope that it will be useful,
''but WITHOUT ANY WARRANTY; without even the implied warranty of
''MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
''GNU General Public License for more details.
''
''You should have received a copy of the GNU General Public License
''along with this program; if not, write to the Free Software
''Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
''
''CONTACT
''
''  phil@phipi.com (First contact may require addition to my whitelist.)
''
''VERSION HISTORY
''
''  0.01 alpha: released 2006.10.21
''  0.02 alpha: released 2006.10.22
''    Fixed a bug that prevented the algorithm from terminating in some cases.
''
''SUMMARY
''
''  This module determines the shortest algorithm for multiplication of a 32-bit operand by a 16-bit fractional constant.
''  The solution is given as a pattern and a mask of the (example) form:
''
''  8800 ++00-000-00+000+
''
''  where the first number (in hex) defines a subproduct to be shifted and added to/subtracted from the multiplicand.
''  The second string is a mask indicating what to do (add, subtract, nop) at each shift position.
''  The instruction counts used to score each solution are based on the following code templates. Which template to use
''  is determined by three things:
''
''     pbits: The number of "one"-bits in the pattern.
''     abits: The number of non-zero operations in the mask.
''     first_neg: 1 if the first non-zero mask position is negative; 0 otherwise.
''
''  Here are the templates, where p0, p1, ..., are the positions from the left (fictional bit 16) of "one"-bits
''  in the pattern; and a0, a1, ..., are the positions from the left (bit 16, again) of non-zero mask bits.
''  (In every solution, the high-order pattern bit will be a "one", so p0 always equals 1.):
''
''    pbits == 1 and abits == 1 and first_neg == 0  Instructions: 1
''
''      sar x,#a0
''
''    pbits == 1 and abits == 1 and first_neg == 1  Instructions: 3
''
''      mov t,x
''      sar t,#a0
''      sub x,t
''
''    pbits == 1 and abits > 1 and first_neg == 0   Instructions: abits * 2
''
''      sar x,#a0
''      mov t,x
''
''      sar t,#a1-a0
''      add/sub x,t             'Second non-zero operation in mask.
''
''      sar t,#a2-a1
''      add/sub x,t             'Third non-zero operation in mask.
''
''      ...
''
''    pbits > 1 and abits == 1 and first_neg == 0   Instructions: pbits * 2
''
''      sar x,#a0
''      mov t,x
''
''      sar t,#p1-p0
''      add x,t
''
''      sar t,#p2-p1
''      add x,t
''
''      ...
''
''    pbits == 1 and abits > 1 and first_neg == 1   Instructions: abits * 2 + 1
''
''      mov t,x
''
''      sar t,#a0
''      sub x,t
''
''      sar t,#a1-a0
''      add/sub x,t             'Second non-zero operation in mask.
''
''      sar t,#a2-a1
''      add/sub x,t             'Third non-zero operation in mask.
''
''      ...
''
''     pbits > 1 and abits == 1 and first_neg == 1  Instructions: pbits * 2 + 1
''
''       mov t,x
''
''       sar t,#a0              'First non-zero operation in mask (a subtract, since first_neg == 1).
''       sub x,t
''
''       sar t,#p1-p0
''       add/sub x,t            'Second non-zero operation in mask.
''
''       ...
''
''    pbits > 1 and abits > 1 and first_neg == 0    Instructions: (pbits - 1) * 2 + (abits - 1) * 2 + 3
''
''       sar x,#a0
''       mov t,x
''
''       sar t,#p1-p0
''       add x,t
''
''       sar t,#p2-p1
''       add x,t
''
''       ...
''
''       mov t,x
''
''       sar t,#a1-a0           'Second non-zero operation in mask.
''       add/sub x,t
''
''       sar t,#a2-a1           'Third non-zero operation in mask.
''       add/sub x,t
''
''       ...
''
''    pbits > 1 and abits > 1 and first_neg == 1    Instructions: (pbits - 1) * 2 + (abits - 1) * 2 + 5
''
''       neg x,x                'Negate first, then do the opposite of what the mask says.
''
''       sar x,#a0
''       mov t,x
''   
''       sar t,#p1-p0
''       add x,t
''
''       sar t,#p2-p1
''       add x,t
''
''       ...
''
''       mov t,x
''
''       sar t,#a1-a0
''       sub/add x,t            'Opposite of second non-zero operation in mask.
''
''       sar t,#a2-a1
''       sub/add x,t            'Opposite of third non-zero operation in mask.
''
''       ...
''
''       neg x,x                'Negate back for final result.
''

CON

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000
  basepin       = 12

  fraction      = $9b75 '<-- THIS IS THE FRACTIONAL MULTIPLIER. Set it to any 16-bit value.
  max_ptn_bits  = 4     'This determines the highest number of one-bits allowed in the pattern.
                        'The lower this number, the faster the search will complete.

  CLS           = $00  'clear screen
  HOME          = $01  'home
  CRSRXY        = $02  'set X, Y position (X, Y follow)
  CRSRLF        = $03  'cursor left
  CRSRRT        = $04  'cursor right
  CRSRUP        = $05  'cursor up
  CRSRDN        = $06  'cursor dn
  USECLR        = $07  'use color C (C follows)
  BS            = $08  'backspace
  TAB           = $09  'tab (8 spaces per)
  LF            = $0A  'linefeed
  CLREOL        = $0B  'clear to end of line
  CLRDN         = $0C  'clear down (to end of window)
  CR            = $0D  'return
  CRSRX         = $0E  'set X position (X follows)
  CRSRY         = $0F  'set Y position (Y follows)
  DEFWIN        = $10  'define window  W (W, Left, Top, nCols, nRows follow)
  USEWIN        = $11  'use window W (W follows)
  DEFCLR        = $12  'define color C (C, FG, BG follow)
  SCRLLF        = $13  'scroll window left
  SCRLRT        = $14  'scroll window right
  SCRLUP        = $15  'scroll window up
  SCRLDN        = $16  'scroll window down
  CHGCLR        = $17  'change all colors in window to C (C follows)
  CLRW          = $1E  'same as CLR, but can be used in strings.
  ESC           = $1F  'escape next character C (i.e. print as-is) (C follows)
  ZERO          = $FF  'can be used for 0, which is not allowed in strings, for command arguments

OBJ

  pr :  "tv_wtext"

VAR

  long  Args[3]

PUB start | res_addr, j, ptn, adsb, abits, pbits, best, first_neg, instr_cnt

  pr.start(basepin)
  Args[0] := 1
  cognew(@begin, @Args)
  repeat while Args[0]
  init(fraction, max_ptn_bits)
  best := 9999
  pr.str(string(CHGCLR, 7, DEFWIN, 1, ZERO, 1, 32, 15))
  pr.hex(fraction, 4)
  pr.out(" ")
  pr.bin(fraction,16)
  pr.str(string(" <- Target", USEWIN, 1))  
  repeat while (res_addr := find_next)      'Get next correct solution until entire space has been searched.
    ptn := long[res_addr]                   'Retrieve the pattern: Sixteen bit pattern to shift and add/sub.
    adsb := long[res_addr][1]               'Retrieve the add/sub mask. 32-bits in reverse order: -+-+-+-+ ... -+-+                                       
    abits~                                  'Initialize bit counts,
    pbits~
    first_neg~                              '  and first_neg
    repeat j from 0 to 15                   'Count the non-zero bits in ptn and adsb.
      if ptn & 1
        pbits++
      ptn >>= 1
      if adsb & 1                           'adsb & 1 => add at this position
        abits++
      elseif adsb & 2                       'adsb & 2 => subtract at this position
        if abits++ == 0                     'Is this the first position?
          first_neg++                       '  Yes: Set first_neg.
      adsb >>= 2
    if pbits == 1 and abits == 1            'Determine which template to use and count the instructions.
      instr_cnt := first_neg << 1 + 1
    elseif pbits == 1 or abits == 1
      instr_cnt := (abits #> pbits) << 1 + first_neg
    else
      instr_cnt := (first_neg + pbits + abits - 2) << 1 + 3
    if instr_cnt < best                     'Is count less than the best count so far?
      best := instr_cnt                     '  Yes: Save new best count.
      ptn := long[res_addr]                 '       Retrieve pattern and mask again,
      adsb := long[res_addr][1]             '         since we've messed with 'em.
      pr.hex(ptn, 4)                        '       Print the new best solution.
      pr.out(" ")
      repeat j from 0 to 15
        if adsb & 1
          pr.out("+")
        elseif adsb & 2
          pr.out("-")
        else
          pr.out("0")
        adsb >>= 2
      pr.out(" ")
      pr.dec(instr_cnt) 
      pr.out(13)
  pr.str(string("Done."))

PUB init(targ, n_ones)

'' Establish the target fraction and maximum number of ones in the pattern.

  Args[1] := targ
  Args[2] := n_ones
  Args[0] := 1
  repeat while Args[0]

PUB search(ptrn, adsb)

'' Begin a new search from the given pattern and mask. (Optional, and not used in this program.)

  Args[1] := ptrn
  Args[2] := adsb
  Args[0] := 2
  repeat while Args[0]

PUB find_next

'' Find the next pattern and mask that yields the correct fraction.

  Args[0] := 3
  repeat while Args[0]
  if Args[1]
    return @Args[1]
  else
    return 0

DAT

begin         mov       go,#0
              wrword    go,par
              
:cmd_lp       rdword    go,par  wz
         if_z jmp       #:cmd_lp
              djnz      go,#:search

:init         mov       ptr,par                 'Entry point for Init.
              add       ptr,#4
              rdlong    target,ptr
              add       ptr,#4
              rdlong    onebits,ptr
              mov       pattern,_0x8000         'Pattern must have msb set, so first pattern is $80000.
              mov       addsub,#0               'Initilaize mask to all nops; it will get incremented first thing.
              call      #fix_ptn                'Superfluous call, unless number of 1-bits must EQUAL onebits.
              jmp       #begin

:search       djnz      go,#:search_lp

              mov       ptr,par                 'Entry Point for Search.
              add       ptr,#4
              rdlong    pattern,ptr
              or        pattern,_0x8000         'Make sure msb of pattern is set,
              and       pattern,_0xffff         '  and top 16 bits are cleared,
              call      #fix_ptn                '  and that number of 1-bits is okay.
              add       ptr,#4
              rdlong    addsub,ptr wz           'Search will start with NEXT mask after the parameter.
              jmp       #begin
              
:search_lp                                      'Entry point for Find.
              add       addsub,#1               'Increment the addsub mask to get to the next one.
                                                'addsub uses two bits for each position:
                                                '  %00 = nop
                                                '  %01 = add
                                                '  %10 = sub
                                                '  %11  is not allowed, so we have to increment past all of them...
              test      addsub,#$03 wz,wc       'Non-zero and no carry (even parity) means %11,
 if_nz_and_nc add       addsub,#$01             '  so increment past it.
              test      addsub,#$0c wz,wc       'Repeat for all bit pairs (loop unrolled for speed)...
 if_nz_and_nc add       addsub,#$04
              test      addsub,#$30 wz,wc
 if_nz_and_nc add       addsub,#$10
              test      addsub,#$c0 wz,wc
 if_nz_and_nc add       addsub,#$40
              test      addsub,_0x0300 wz,wc
 if_nz_and_nc add       addsub,#$100
              test      addsub,_0x0c00 wz,wc
 if_nz_and_nc add       addsub,_0x0400
              test      addsub,_0x3000 wz,wc
 if_nz_and_nc add       addsub,_0x1000
              test      addsub,_0xc000 wz,wc
 if_nz_and_nc add       addsub,_0x4000
              test      addsub,_0x00030000 wz,wc
 if_nz_and_nc add       addsub,_0x00010000
              test      addsub,_0x000c0000 wz,wc
 if_nz_and_nc add       addsub,_0x00040000
              test      addsub,_0x00300000 wz,wc
 if_nz_and_nc add       addsub,_0x00100000
              test      addsub,_0x00c00000 wz,wc
 if_nz_and_nc add       addsub,_0x00400000
              test      addsub,_0x03000000 wz,wc
 if_nz_and_nc add       addsub,_0x01000000
              test      addsub,_0x0c000000 wz,wc
 if_nz_and_nc add       addsub,_0x04000000
              test      addsub,_0x30000000 wz,wc
 if_nz_and_nc add       addsub,_0x10000000
              test      addsub,_0xc0000000 wz,wc
 if_nz_and_nc add       addsub,_0x40000000 wz
              test      addsub,addsub wz        'Is the final result zero?
         if_z mov       addsub,#1               '  Yes: We've checked all the masks for this pattern: Start over.
        if_nz jmp       #:test                  '  No:  Test the new pattern/mask combo.

:next_ptn     call      #inc_ptn                '  Yes: Get the next pattern.
         if_z jmp       #:found                 '       Done if zero flag is set.              

:test         mov       shiftp,pattern          'Copy pattern and mask into shift registers.
              mov       shifta,addsub
              mov       answer,#0               'Clear the answer.

:test_lp      shr       shifta,#1 wc            'Test last two bits of mask,
              test      shifta,#1 wz
              shr       shifta,#1               '  and finish shifting them out.
         if_c add       answer,shiftp           'If lsb was 1, then add the pattern to the answer.
        if_nz sub       answer,shiftp           'If 2nd-to-lsb was 1, then subtract the pattern from the answer.
              shr       shiftp,#1 wc            'Shift pattern to the right,
        if_nc jmp       #:test_lp               '  but not beyond where rightmost one-bit lands in lsb.

              test      shifta,shifta wz        'We've finished shifting the pattern. Are there ops remaining in mask?
        if_nz mov       addsub,#1               '  Yes: Invalid combination, so start mask over,
        if_nz jmp       #:next_ptn              '         and get next pattern.               

              and       answer,_0xffff          '  No:  Only interested in lower 16 bits.
              cmp       answer,target wz        '       Did we get the right answer?
        if_nz jmp       #:search_lp             '         No:  Go back and try the next combo.              

:found        mov       ptr,par                 '         Yes: Return the pattern and mask.
              add       ptr,#4
              wrlong    pattern,ptr
              add       ptr,#4
              wrlong    addsub,ptr
              jmp       #begin
              

inc_ptn       add       pattern,#1              'Increment pattern to get the next one.
              and       pattern,_0xffff wz      'Is it zero?
         if_z jmp       inc_ptn_ret             '  Yes: No more new patterns to check.

fix_ptn       mov       shiftp,pattern          '  No:  Count the 1-bits in pattern.
              mov       count,#0
:chk_lp       shr       shiftp,#1 wz,wc
         if_c add       count,#1
        if_nz jmp       #:chk_lp
              cmp       count,onebits wz,wc     '       Does the count exceed the max?
 if_nz_and_nc jmp       #inc_ptn                '         Yes: Get the next one.
              test      pattern,pattern wz      '         No:  Clear the zero flag. (We know pattern <> 0.)
inc_ptn_ret
fix_ptn_ret   ret                               'Return with the new pattern or the zero flag set.


_0xffff       long      $ffff
_0x8000       long      $8000

_0x0300       long      $0300
_0x0c00       long      $0c00
_0x3000       long      $3000
_0xc000       long      $c000
_0x00030000   long      $00030000
_0x000c0000   long      $000c0000
_0x00300000   long      $00300000
_0x00c00000   long      $00c00000
_0x03000000   long      $03000000
_0x0c000000   long      $0c000000
_0x30000000   long      $30000000
_0xc0000000   long      $c0000000

_0x0400       long      $0400
_0x1000       long      $1000
_0x4000       long      $4000
_0x00010000   long      $00010000
_0x00040000   long      $00040000
_0x00100000   long      $00100000
_0x00400000   long      $00400000
_0x01000000   long      $01000000
_0x04000000   long      $04000000
_0x10000000   long      $10000000
_0x40000000   long      $40000000

_0x80000000   long      $80000000
_0xffffffff   long      $ffffffff

go            res       1
ptr           res       1
target        res       1
onebits       res       1
pattern       res       1
addsub        res       1
count         res       1
shiftp        res       1
shifta        res       1
answer        res       1